"""
A Qt widget that displays a SVG based synoptic view.
It allows navigation in the form of zooming, panning and clicking
various areas to zoom in.
"""

import logging
import os
import json
from pathlib import Path

from PyQt5 import Qt, QtCore, QtWidgets
from PyQt5.QtWebChannel import QWebChannel
from PyQt5.QtWebEngineWidgets import QWebEnginePage, QWebEngineView

from .jsinterface import JSInterface


class LoggingWebPage(QWebEnginePage):
    """
    Use a Python logger to print javascript console messages.
    Very useful for debugging javascript...
    """

    def __init__(self, logger=None, parent=None):
        super().__init__(parent)
        if not logger:
            logger = logging
        self.logger = logger

    def javaScriptConsoleMessage(self, source_id, msg, line_number, filename):
        # don't use the logger for now; too verbose :)
        print(f"JsConsole({source_id}:{line_number}):\n\t{msg}\n\t{filename}")


class SynopticWidget(QtWidgets.QWidget):
    """
    A Qt widget displaying a SVG synoptic in a webview.

    This widget does not have a "backend", so it's intended
    to be subclassed to add control system specific behavior.
    See TaurusSynopticWidget for a TANGO implementation.
    """

    subscribe = QtCore.pyqtSignal(str)

    def __init__(self, url=None, parent=None, *args, **kwargs):
        super().__init__(parent)
        self.subscribe.connect(self._handle_subscriptions)
        self._url = url
        self._setup_ui(url)
        self._modelNames = None

    def _setup_ui(self, url=None, section=None):
        self.hbox = QtWidgets.QHBoxLayout(self)
        self.hbox.setContentsMargins(0, 0, 0, 0)
        self.hbox.layout().setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.hbox)
        if url:
            self.set_url(url, section)

    def set_url(self, url, section=None):
        # TODO: probably breaks things if the url is already set
        self._url = url
        self.splitter = QtWidgets.QSplitter(self)
        self.splitter.setOrientation(QtCore.Qt.Vertical)
        self.hbox.addWidget(self.splitter)
        view = self._create_view(url, section)
        self.splitter.addWidget(view)

    def setConfig(self, config_file):
        # build a javascript defining the models
        text = "var modelNames ={"
        with open(config_file) as read_file:
            data = json.load(read_file)
            for key in data.keys():
                text += key + ' : "' + data[key] + '",'
        text += "};"
        self._modelNames = text

    def _create_view(self, html=None, section=None):
        """
        Create the webview that will display the synoptic itself
        """
        view = QWebEngineView(self)
        channel = QWebChannel(self)
        # This is supposedly an optimisation. Disable if there are
        # graphical artifacts or something.
        view.settings().TiledBackingStoreEnabled = True

        page = LoggingWebPage()
        view.setPage(page)
        view.setContextMenuPolicy(QtCore.Qt.PreventContextMenu)

        # setup the JS interface
        frame = view.page()
        self.js = JSInterface(frame)
        self.js.subscription.connect(self.subscribe)
        channel.registerObject("QtBackend", self.js)
        frame.setWebChannel(channel)

        # mouse interaction signals
        self.clicked = self.js.leftclicked
        self.rightClicked = self.js.rightclicked
        self.hovered = self.js.hovered
        self.clicked.connect(self._on_click)
        self.rightClicked.connect(self._on_rightclick)
        self.hovered.connect(self._on_hover)

        # load the page
        # need to set the "base URL" for the webview to find the
        # resources (js, css).
        base_url = QtCore.QUrl().fromLocalFile(
            str(Path(__file__).resolve().parent) + "/web/"
        )

        # some ugly magic to get the path to the SVG file right. It
        # needs to be absolute because local paths go to the base URL.
        html_folder = Path(html).resolve().parent

        with open(html) as f:
            # TODO: use template
            text = f.read().replace("${path}", html_folder.as_uri())
            if self._modelNames is not None:
                if "/*configplaceholder*/" in text:
                    texta, textb = text.split("/*configplaceholder*/", 1)
                    text = texta + self._modelNames + textb
            view.setHtml(text, base_url)

        self.view = view
        return view

    def _handle_subscriptions(self, models):
        # we get the subscribed models as a comma separated list from Qt,
        # let's deliver something neater.
        if models:
            self.handle_subscriptions(str(models).split("\n"))
        else:
            self.handle_subscriptions([])

    def handle_subscriptions(self, models):
        # This noop needs to be overridden in order for subscriptions
        # to actually do anything!
        # "models" is a list of models that are currently visible
        # in the synoptic.
        pass

    def _on_click(self, kind, name):
        # this is a workaround for a strange behavior; under some circumstances
        # we get QStrings here, and sometimes unicode. Investigate!
        self.on_click(str(kind), str(name))

    def on_click(self, kind, name):
        """
        Default behavior on click. Override to change!
        """
        if kind == "section":
            self.zoom_to(kind, name)
        elif kind == "model":
            self.select(kind, [name])

    def _on_rightclick(self, kind, name):
        # see _on_click()
        self.on_rightclick(str(kind), str(name))

    def on_rightclick(self, kind, name):
        """
        Placeholder; override me!
        """
        pass

    def _on_hover(self, section, models):
        # we get a comma separated string of concatenated models here.
        splitmodels = str(models).split("\n") if models else []
        self.on_hover(section, splitmodels)

    def on_hover(self, section, models):
        """
        Show a basic 'tooltip' when the mouse pointer is over an item.
        """
        # Override this to add more interesting behavior
        if section:
            self.js.evaluate("synoptic.showTooltip();")
            self.js.evaluate(f"synoptic.setTooltipHTML('{section}')")
        elif models:
            self.js.evaluate("synoptic.showTooltip();")
            self.js.evaluate(f"synoptic.setTooltipHTML('{models[0]}')")
        else:
            self.js.evaluate("synoptic.hideTooltip();")

    def keyPressEvent(self, e):
        super().keyPressEvent(e)
        # show inspector window
        if e.key() == QtCore.Qt.Key_F12:
            DEBUG_PORT = "5588"
            DEBUG_URL = f"[http://127.0.0.1](http://127.0.0.1/):{DEBUG_PORT}"
            os.environ["QTWEBENGINE_REMOTE_DEBUGGING"] = DEBUG_PORT
            self.inspector = QWebEngineView()
            self.inspector.setWindowTitle("Web Inspector")
            self.inspector.load(QtCore.QUrl(DEBUG_URL))
            self.view.page().setDevToolsPage(self.inspector.page())
            self.inspector.show()

    # # # 'Public' API # # #

    def zoom_to(self, kind, name):
        """
        Move the view so that the given object is visible
        """
        self.js.evaluate(f"synoptic.zoomTo({str(kind)!r}, {str(name)!r})")

    def select(self, kind, names, replace=True):
        """
        Set a list of items as 'selected'. By default unselects all
        previously selected things first.
        """
        if replace:
            self.js.evaluate("synoptic.unselectAll()")
        if names:
            for name in names:
                self.js.evaluate(f"synoptic.select({str(kind)!r}, {[str(name)]!r})")

    def unselect_all(self):
        self.js.evaluate("synoptic.unselectAll();")

    # def send_debug_message(self, msg):
    #     self.js.evaluate("Tango.debugMessage(%r)" % msg)


if __name__ == "__main__":
    import sys

    qapp = Qt.QApplication([])
    sw = SynopticWidget(sys.argv[1])
    sw.show()
    qapp.exec_()



from taurus.qt.qtgui.panel import TaurusForm
from taurus.qt.qtgui.container import TaurusWidget
from taurus.external.qt import Qt, QtGui, QtCore
from taurus.qt.qtgui.button import TaurusCommandButton
from taurus.qt.qtgui.display import TaurusLabel
import tango
import taurus


class TaurusFormExpanding(TaurusForm):
    """A TaurusForm that adapts its size to its children"""

    max_vertical_size = 500

    def sizeHint(self):
        frame = self.scrollArea.widget()
        framehint = frame.sizeHint()
        hint = Qt.QSize(
            framehint.width(), min(self.max_vertical_size, framehint.height())
        )
        return hint + self.buttonBox.sizeHint() + Qt.QSize(0, 20)


class StateComposerWidget(TaurusFormExpanding):
    """A widget for displaying devices listed in a StateComposer

    StateComposer is a Facade device that combines some states from
    a list of devices
    """

    _customWidgetMap = getattr(
        taurus.tauruscustomsettings, "T_FORM_CUSTOM_WIDGET_MAP", {}
    )

    def setModel(self, model):
        if model:
            # Get the devices from the device states property
            # (returns an array of "device/State" attributes)
            states = tango.Database().get_device_property(model, "states")["states"]
            # Remove the "State" attribute to get the device name
            devices = [state.rsplit("/", 1)[0] for state in states]
            self.setCustomWidgetMap(
                getattr(taurus.tauruscustomsettings, "T_FORM_CUSTOM_WIDGET_MAP", {})
            )
            self.setWithButtons(False)
            TaurusFormExpanding.setModel(self, devices)


class CommandsWidget(TaurusWidget):
    """
    Simple widget that shows a number of Command buttons.
    """

    def __init__(self, parent=None, commands=None):
        super().__init__(parent)
        self.commands = commands
        self.buttons = {}
        self.setupUi()

    def setupUi(self):
        w = TaurusWidget()
        self.setLayout(Qt.QHBoxLayout())
        self.layout().addWidget(w)

        self.modelLabel = QtGui.QLabel(w)
        self.layout().addWidget(self.modelLabel)

        self.stateLabel = TaurusLabel(w)
        self.layout().addWidget(self.stateLabel)

        for command, _ in self.commands:
            button = TaurusCommandButton(w, command=command)
            self.layout().addWidget(button)
            self.buttons[command] = button

    def setModel(self, model):
        super().setModel(model)
        self.setWindowTitle(str(model))
        self.modelLabel.setText(str(model))
        self.stateLabel.setModel(f"{model}/state")
        self.stateLabel.setBgRole("state")
        self.stateLabel.setFgRole("state")
        for button in self.buttons.values():
            button.setModel(str(model))


class CommandsWidgetPopup(CommandsWidget):
    """
    Popup variant of CommandsWidget.
    It moves to mouse on start and it closes on unfocus
    or after successful command.
    On show it hides irrelevant buttons (based on the state).
    It also closes automatically after 10 seconds.

    This class shouldn't be used directly.
    Define your own class with "commands" as class attribute
    """

    def __init__(self, parent=None):
        super().__init__(parent, self.commands)
        self.setWindowFlags(QtCore.Qt.FramelessWindowHint)
        self.installEventFilter(self)
        for button in self.buttons.values():
            button.clicked.connect(self.close)

    def eventFilter(self, object, event):
        """
        Event filter to close if no focus
        """
        if event.type() == QtCore.QEvent.WindowDeactivate:
            self.close()
        return super().eventFilter(object, event)

    def show(self):
        super().show()
        self.showhidebuttons()
        self.move_to_mouse()
        QtCore.QTimer.singleShot(10000, lambda: self.close())

    def showhidebuttons(self):
        dev = taurus.Device(self.model)
        for command, state in self.commands:
            button = self.buttons[command]
            if dev.State() == state:
                button.hide()
            else:
                button.show()
        self.adjustSize()

    def move_to_mouse(self):
        """
        Moves the window to the mouse cursor
        """
        try:
            pos = QtGui.QCursor.pos()
            self.move(
                round(pos.x() - 4 * (self.width() / 5)),
                round(pos.y() - (self.height() / 2)),
            )
        except Exception:
            pass


class ValvePopup(CommandsWidgetPopup):
    """Quick open/close for vacuum valves"""

    commands = (("Open", tango.DevState.OPEN), ("Close", tango.DevState.CLOSE))


class PlcResetPopup(CommandsWidgetPopup):
    """Variant of CommandsWidgetPopup with a single command
    to reset a PLC.

    This class doesn't use a tango command but writes a value
    to a specific attribute.
    """

    commands = (("Reset", None),)

    def __init__(self, parent=None):
        self.command = self.commands[0][0]
        super().__init__(parent)

    def setupUi(self):
        w = TaurusWidget()
        self.setLayout(Qt.QHBoxLayout())
        self.layout().addWidget(w)

        self.modelLabel = QtGui.QLabel(w)
        self.layout().addWidget(self.modelLabel)

        self.stateLabel = TaurusLabel(w)
        self.layout().addWidget(self.stateLabel)

        button = QtGui.QPushButton(self.command, self)
        button.clicked.connect(self.doReset)

        self.layout().addWidget(button)
        self.buttons[self.command] = button

    def doReset(self):
        print("Resetting PLC")
        self.resetAttr.write(1)

    def setModel(self, model):
        TaurusWidget.setModel(self, model)
        self.setWindowTitle(str(model))
        self.modelLabel.setText(str(model))
        self.stateLabel.setModel(f"{model}/state")
        self.stateLabel.setBgRole("rvalue")
        try:
            proxy = tango.DeviceProxy(model)
            attributes = proxy.get_attribute_list()
            # Depending on beamlines
            # reset attribute can be BXXXX_VAC_RESET_C or BXXXX_VAC_PLC01_RESET_C
            parts = model.split("/")
            reset_attributes = [
                attribute
                for attribute in attributes
                if attribute.upper().endswith("_RESET_C")
                and attribute.upper().startswith(
                    f"{parts[0].upper()}_{parts[1].upper()}_"
                )
            ]
            if len(reset_attributes) == 1:
                attr_name = reset_attributes[0]
                self.resetAttr = tango.AttributeProxy(f"{model}/{attr_name}")
                print(f"PLC reset attribute: {model}/{attr_name}")
            else:
                print(
                    f"Couldn't find reset attribute name... {attributes=} {reset_attributes=}"
                )
        except Exception as e:
            print(e)



"""
A simple commandline way to test a synoptic file
"""

import json
import os
import sys
from optparse import OptionParser
from string import Template
from tempfile import NamedTemporaryFile

from taurus.qt.qtgui.application import TaurusApplication

from svgsynoptic2.taurussynopticwidget import TaurusSynopticWidget


def main():

    parser = OptionParser("usage: %prog [options] SVGFILE")
    parser.add_option(
        "-s",
        "--size",
        dest="size",
        help="Window size on form WIDTH,HEIGHT",
        metavar="WINSIZE",
    )
    parser.add_option(
        "-t", "--title", dest="title", help="Window title", metavar="WINTITLE"
    )
    parser.add_option(
        "-z",
        "--zoomsteps",
        dest="zoomsteps",
        metavar="ZOOMSTEPS",
        help="Zoom levels, on form ZOOM1,ZOOM2,...",
        default="1",
    )

    app = TaurusApplication(cmd_line_parser=parser)

    args = app.get_command_line_args()
    if len(args) != 1:
        sys.exit("You need to specify the SVG file to load!")
    svg = args[0]
    options = app.get_command_line_options()

    widget = TaurusSynopticWidget()

    # We'd like the synoptic to "select" the relevant item when
    # the user focuses on a panel. Let's connect a handler to
    # the focusChanged signal that does this.
    def onfocus(old, new):
        if new and hasattr(new, "window"):
            for device, panel in widget._panels.items():
                if panel == new.window():
                    widget.select("model", [device])

    app.focusChanged.connect(onfocus)

    # need absolute path to the SVG file
    svgfile = os.path.abspath(svg)

    # since the svg currently needs to be hardcoded in the HTML, we
    # create a temporary HTML file from a static template.
    path = os.path.dirname(__file__)
    template = os.path.join(path, "web", "template.html")
    with open(template) as f:
        tmpl = Template(f.read())
    zoomsteps = [int(z) for z in options.zoomsteps.split(",")]
    config = {"view": {"zoomSteps": zoomsteps}}
    html = tmpl.substitute(path="/web", svgfile=svgfile, config=json.dumps(config))
    with NamedTemporaryFile(suffix=".html") as tf:
        tf.write(html)
        tf.flush()
        widget.setModel(tf.name)
        if options.size:
            w, h = options.size.split(",")
            widget.resize(int(w), int(h))

        widget.setWindowTitle(options.title or os.path.basename(svg))
        widget.show()
        app.exec_()


if __name__ == "__main__":
    main()
